use Module::Build;

my $b = Module::Build->new
 (module_name   => 'Data::Table::Text',
  dist_author   => 'philiprbrenan@gmail.com',
  dist_abstract => 'Dump data in tabular text format',
  license       => 'perl',
  perl          => '5.16',
  requires      => {Data::Dump=>0, File::Path=>0, File::Glob=>0},
  configure_requires => { 'Module::Build' => 0.42 },
  create_readme =>  1,
 );

$b->create_build_script();
